/**
 * 
 */
package de.mpc.prolic.comparison;

import java.util.ArrayList;
import java.util.ResourceBundle;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.mpc.prolic.model.CleavageSite;
import de.mpc.prolic.model.Enzyme;
import de.mpc.prolic.model.Peptide;
import de.mpc.prolic.model.Protein;

/**
 * This class provides different static methods for comparison of the peptide sets of two proteins
 * @author Michael Kohl
 *
 */

public class PeptideBasedComparison {

	private final static ResourceBundle bundleMess = ResourceBundle.getBundle("de.mpc.prolic.resources.MessageBundle");		//A localized bundle for user information and for logging
	private static final Logger log4j = LogManager.getLogger(PeptideBasedComparison.class.getName());

	/**
	 * Performs a peptide based comparison that considers only the available peptide sets of the two proteins specified. 
	 * Two proteins are different, if each peptide set of the compared proteins contains at least one peptide that is not part of the peptide
	 * set of the other protein 
	 * @param aProt
	 * @param anOtherProt
	 * @return a boolean indicating whether two proteins are considered similar with respect to a peptide based comparison that considers only the available peptide sets of two given proteins.
	 */
	public static boolean bilateralCompPeptOnly(Protein aProt, Protein anOtherProt){
		boolean isSimilar=true;
		boolean aProtHasUniquePepts=false;								//Stores the information, whether there are measured peptides for aProt that are not exist in the protein sequence of anOtherProt
		boolean anOtherProtHasUniquePepts=false;						//Stores the information, whether there are measured peptides for anOtherProt that are not exist in the protein sequence of aProt
		boolean peptideMatchesOccurred=false;							//Stores the information, whether there are is a match between the peptide under consideration with any other peptide in the peptide 
																		//set of the compared protein.
		ArrayList <Peptide> aProtPept=aProt.getPeptides();				//All the peptides from aProt
		ArrayList <Peptide> anOtherProtPept=anOtherProt.getPeptides(); 	//All the peptides from anOtherProt
		Peptide aPept;
		Peptide aPeptToCompare;
		//Considering the peptides of aProt
		for (int i=0;i<aProtPept.size();i++){
			aPept=aProtPept.get(i);
			peptideMatchesOccurred=false;
			for (int j=0;j<anOtherProtPept.size();j++){
				aPeptToCompare=anOtherProtPept.get(j);
				if(aPept.getSequence().equals(aPeptToCompare.getSequence())){
					peptideMatchesOccurred=true;
				}
			}
			if(!peptideMatchesOccurred){
				aProtHasUniquePepts=true;
			}
		}
		//Considering the Peptides of anOtherProt
		for (int i=0;i<anOtherProtPept.size();i++){
			aPept=anOtherProtPept.get(i);
			peptideMatchesOccurred=false;
			for (int j=0;j<aProtPept.size();j++){
				aPeptToCompare=aProtPept.get(j);
				if(aPept.getSequence().equals(aPeptToCompare.getSequence())){
					peptideMatchesOccurred=true;
				}
			}
			if(!peptideMatchesOccurred){
				anOtherProtHasUniquePepts=true;
			}
		}
		if(aProtHasUniquePepts && anOtherProtHasUniquePepts){
			isSimilar=false;
		}
		return isSimilar;
	}
	
	
	/**
	 * Performs a peptide based comparison of two proteins obtained from different protein list. Similarity requires that
	 * either all peptides obtained from one protein are identified in the compared protein or that for the missing identified peptide sequences a
	 * a corresponding substring exists in the compared protein sequence. 
	 * As a result, a sufficient criteria for the presence of a peptide in the compared protein is checking of occurrence as a substring in 
	 * the protein sequence.
	 * @param aProt
	 * @param anOtherProt
	 * @param enzymes
	 * @return a boolean indicating whether two proteins are considered similar with respect to a peptide based comparison that takes also the protein sequence into account.
	 */
	public static boolean bilateralCompPeptSeq(Protein aProt, Protein anOtherProt, ArrayList<Enzyme> enzymes){
		Enzyme e = enzymes.get(0); 										//Currently only a single enzyme definition is allowed.
		ArrayList <CleavageSite> cs = e.getCleavageSites();
		CleavageSite aCs;
		boolean isSimilar=true;
		boolean aProtHasUniquePepts=false;								//Stores the information, whether there are measured peptides for aProt that are not exist in the protein sequence of anOtherProt
		boolean anOtherProtHasUniquePepts=false;						//Stores the information, whether there are measured peptides for anOtherProt that are not exist in the protein sequence of aProt
		//TODO: Reducing of the code: terminalPeptUniqueness and regularPeptUniqueness are not necessary. If either terminalPeptUniqueness or regularPeptUniqueness is reached the loops can be terminated
		boolean terminalPeptUniqueness;
		boolean regularPeptUniqueness;
		boolean currentPeptUniqueness;	
		boolean cleavageSiteDepUniqueness;
		ArrayList <Peptide> aProtPept=aProt.getPeptides();				//all the peptides from aProt
		ArrayList <Peptide> anOtherProtPept=anOtherProt.getPeptides(); 	//all the peptides from anOtherProt
		Peptide aPept;
		//String aPeptSeqMod;											//the original peptide sequence is modified in order to account for adjacent cleavage sites in the protein sequence.
		String aProtSeq=aProt.getSequence();
		String aProtCTerminalPeptSeq;
		String aProtNTerminalPeptSeq;
		String anOtherProtSeq=anOtherProt.getSequence();
		String anOtherProtCTerminalPeptSeq;
		String anOtherProtNTerminalPeptSeq;
		/**
		 * The algorithm requires special treatment of the terminal peptides, which is also dependent on the cleavage of the enzyme:
		 * Case 1 - C- terminal cleavage:
		 * a) requires special consideration of the n-terminal peptide, because in this situation it is not necessary to ADD cleavage sites to the n-terminal
		 * side of the peptide for determination of peptide uniqueness.
		 * b) requires special consideration of the c-terminal peptide, because in this situation it is not necessary that there IS a cleavage site at the 
		 * c - terminal end of the peptide.
		 * Case 2 - N -Terminal cleavage:
		 * Considering the n-terminal cleavage the situation analogously reversed.
		 */
		//Calculation of terminal peptides
		aProtCTerminalPeptSeq=calcC_TerminalPeptide(aProtSeq, e.getRegExDigestion());
		anOtherProtCTerminalPeptSeq=calcC_TerminalPeptide(anOtherProtSeq, e.getRegExDigestion());
		aProtNTerminalPeptSeq=calcN_TerminalPeptide(aProtSeq, e.getRegExDigestion());
		anOtherProtNTerminalPeptSeq=calcN_TerminalPeptide(anOtherProtSeq, e.getRegExDigestion());
		//Considering the Peptides of aProt
		for (int i=0;i<aProtPept.size();i++){
			aPept=aProtPept.get(i);
			currentPeptUniqueness=false;
			regularPeptUniqueness=true;
			terminalPeptUniqueness=false;
			if((aPept.getSequence().equals(aProtCTerminalPeptSeq))||(aPept.getSequence().equals(aProtNTerminalPeptSeq))){//Special case: the current peptide in the peptide set 
				//is a terminal peptide of this protein. This requires comparison with the corresponding terminal peptide in the other protein.
				if(aPept.getSequence().equals(aProtCTerminalPeptSeq)){
					terminalPeptUniqueness=determineUniquenessOfTerminalPeptide(aPept.getSequence(),anOtherProtSeq,false, e.getRegExDigestion());//compare with the c-terminal side
				}
				if(aPept.getSequence().equals(aProtNTerminalPeptSeq)){
					terminalPeptUniqueness=determineUniquenessOfTerminalPeptide(aPept.getSequence(),anOtherProtSeq,true, e.getRegExDigestion());//compare with the n-terminal side
				}
				if(terminalPeptUniqueness){
					currentPeptUniqueness=true;
				}
			}else{//'regular' peptide
				for(int j=0; j<cs.size();j++){
					aCs=cs.get(j);
					cleavageSiteDepUniqueness=determinePeptideUniqueness(aPept, anOtherProtSeq, aCs, e.getPerformsN_terminalCleavage());
					if (!cleavageSiteDepUniqueness){
						/* regularPeptUniqueness starts in condition true, but if there is a single sequence + cleavage site combination which is existing in the protein sequence of the compared
						 * protein this results in the non-uniqueness of the considered regular peptide (regularPeptUniqueness=false). 
						 */
						regularPeptUniqueness=false;
					}
				}
				if(regularPeptUniqueness){
					currentPeptUniqueness=true;
				}
			}
			if(currentPeptUniqueness){//a single occurence of a unique peptide is sufficient
				aProtHasUniquePepts=true;
			}
		}
		//Considering the peptides of anOtherProt
		for (int i=0;i<anOtherProtPept.size();i++){
			aPept=anOtherProtPept.get(i);
			currentPeptUniqueness=false;
			regularPeptUniqueness=true;
			terminalPeptUniqueness=false;
			if((aPept.getSequence().equals(anOtherProtCTerminalPeptSeq))||(aPept.getSequence().equals(anOtherProtNTerminalPeptSeq))){//Special case: the current peptide in the peptide set is the terminal peptide of this protein. 
				if(aPept.getSequence().equals(anOtherProtCTerminalPeptSeq)){
					terminalPeptUniqueness=determineUniquenessOfTerminalPeptide(aPept.getSequence(),aProtSeq, false, e.getRegExDigestion());
				}
				if(aPept.getSequence().equals(anOtherProtNTerminalPeptSeq)){
					terminalPeptUniqueness=determineUniquenessOfTerminalPeptide(aPept.getSequence(),aProtSeq, true, e.getRegExDigestion());
				}				
				if(terminalPeptUniqueness){
					currentPeptUniqueness=true;
				}
			}else{//'regular' peptide
				for(int j=0; j<cs.size();j++){
					aCs=cs.get(j);
					cleavageSiteDepUniqueness=determinePeptideUniqueness(aPept, aProtSeq, aCs, e.getPerformsN_terminalCleavage());
					if (!cleavageSiteDepUniqueness){
						regularPeptUniqueness=false;
					}
				}
				if(regularPeptUniqueness){
					currentPeptUniqueness=true;
				}
			}
			if(currentPeptUniqueness){
				anOtherProtHasUniquePepts=true;
			}
		}
		if(aProtHasUniquePepts && anOtherProtHasUniquePepts){//sufficient requirement for dissimilarity is existence of at least on unique peptide in a protein
			isSimilar=false;
		}
		return isSimilar;
	}
		
	/**
	 * Performs a peptide based comparison of two proteins obtained from different protein lists. Similarity requires that
	 * either all peptides obtained from one protein are identified in the compared protein or that for the missing identified peptide sequences a
	 * a corresponding substring exists in the compared protein sequence. 
	 * This method considers modifications of amino acid residues.
	 * As a result, a sufficient criteria for the presence of a peptide in the compared protein is checking of occurrence as a substring in 
	 * the protein sequence.
	 * @param aProt
	 * @param anOtherProt
	 * @param enzymes
	 * @return a boolean indicating whether two proteins are considered similar with respect to a peptide based comparison that takes also both the protein sequence and existing post translational modifications into account.
	 */
	public static boolean bilateralCompPeptSeqMod(Protein aProt, Protein anOtherProt, ArrayList<Enzyme> enzymes){
		Enzyme e = enzymes.get(0); 										//Currently only a single enzyme definition is allowed.
		boolean isSimilar=true;
		boolean aProtHasUniquePepts=false;								//Stores the information, whether there are measured peptides for aProt that are not exist in the protein sequence of anOtherProt
		boolean anOtherProtHasUniquePepts=false;						//Stores the information, whether there are measured peptides for anOtherProt that are not exist in the protein sequence of aProt
		//TODO: Reducing of the code: terminalPeptUniqueness and regularPeptUniqueness are not necessary. If either terminalPeptUniqueness or regularPeptUniqueness is reached the loops can be terminated
		boolean terminalPeptUniqueness;
		boolean regularPeptUniqueness;
		boolean currentPeptUniqueness;	
		boolean cleavageSiteDepUniqueness;
		ArrayList <Peptide> aProtPept=aProt.getPeptides();				//all the peptides from aProt
		ArrayList <Peptide> anOtherProtPept=anOtherProt.getPeptides(); 	//all the peptides from anOtherProt
		Peptide aPept;
		//String aPeptSeqMod;											//the original peptide sequence is modified in order to account for adjacent cleavage sites in the protein sequence.
		String aProtSeq=aProt.getSequence();
		String aProtCTerminalPeptSeq;
		String aProtNTerminalPeptSeq;
		String anOtherProtSeq=anOtherProt.getSequence();
		String anOtherProtCTerminalPeptSeq;
		String anOtherProtNTerminalPeptSeq;
		/**
		 * The algorithm requires special treatment of the terminal peptides, which is also dependent on the cleavage cleavage of the enzyme:
		 * Case 1 - C- terminal cleavage:
		 * a) requires special consideration of the n-terminal peptide, because in this situation it is not necessary to ADD cleavage sites to the n-terminal
		 * side of the peptide for determination of peptide uniqueness.
		 * b) requires special consideration of the c-terminal peptide, because in this situation it is not necessary that there IS a cleavage site at the 
		 * c - terminal end of the peptide.
		 * Case 2 - N -Terminal cleavage:
		 * Considering the n-terminal cleavage the situation analogously reversed.
		 */
		//Calculation of terminal peptides
		aProtCTerminalPeptSeq=calcC_TerminalPeptide(aProtSeq, e.getRegExDigestion());
		anOtherProtCTerminalPeptSeq=calcC_TerminalPeptide(anOtherProtSeq, e.getRegExDigestion());
		aProtNTerminalPeptSeq=calcN_TerminalPeptide(aProtSeq, e.getRegExDigestion());
		anOtherProtNTerminalPeptSeq=calcN_TerminalPeptide(anOtherProtSeq, e.getRegExDigestion());
		//Considering the Peptides of aProt
		for (int i=0;i<aProtPept.size();i++){
			aPept=aProtPept.get(i);
			currentPeptUniqueness=false;
			regularPeptUniqueness=true;
			terminalPeptUniqueness=false;
			if((aPept.getSequence().equals(aProtCTerminalPeptSeq))||(aPept.getSequence().equals(aProtNTerminalPeptSeq))){//Special case: the current peptide in the peptide set 
				//is a terminal peptide of this protein. This requires comparison with the corresponding terminal peptide in the other protein.
				if(aPept.getSequence().equals(aProtCTerminalPeptSeq)){
					terminalPeptUniqueness=determineUniquenessOfTerminalPeptide(aPept.getSequence(),anOtherProtSeq,false,e.getRegExDigestion());//compare with the c-terminal side
				}
				if(aPept.getSequence().equals(aProtNTerminalPeptSeq)){
					terminalPeptUniqueness=determineUniquenessOfTerminalPeptide(aPept.getSequence(),anOtherProtSeq,true,e.getRegExDigestion());//compare with the n-terminal side
				}
				if(terminalPeptUniqueness){//a single event of terminalPeptUniqueness==true is sufficient to define the aProtHasUniquePepts status to true 
					currentPeptUniqueness=true;
				}				
			}else{//'regular' peptide
				cleavageSiteDepUniqueness=determinePeptideUniquenessMod(aPept, anOtherProtSeq, e.getRegExDigestion());
				if (!cleavageSiteDepUniqueness){
					/* regularPeptUniqueness starts in condition true, but if there is a single sequence + cleavage site combination which is existing in the protein sequence of the compared
					 * protein this results in the non-uniqueness of the considered regular peptide (regularPeptUniqueness=false). 
					 */
					regularPeptUniqueness=false;
				}
				if(regularPeptUniqueness){
					currentPeptUniqueness=true;
				}
			}
			if(currentPeptUniqueness){//a single occurence of a unique peptide is sufficient
				aProtHasUniquePepts=true;
			}
		}
		//Considering the peptides of anOtherProt
		for (int i=0;i<anOtherProtPept.size();i++){
			aPept=anOtherProtPept.get(i);
			currentPeptUniqueness=false;
			regularPeptUniqueness=true;
			terminalPeptUniqueness=false;
			if((aPept.getSequence().equals(anOtherProtCTerminalPeptSeq))||(aPept.getSequence().equals(anOtherProtNTerminalPeptSeq))){//Special case: the current peptide in the peptide set is the terminal peptide of this protein. 
				if(aPept.getSequence().equals(anOtherProtCTerminalPeptSeq)){
					terminalPeptUniqueness=determineUniquenessOfTerminalPeptide(aPept.getSequence(),aProtSeq, false, e.getRegExDigestion());
				}
				if(aPept.getSequence().equals(anOtherProtNTerminalPeptSeq)){
					terminalPeptUniqueness=determineUniquenessOfTerminalPeptide(aPept.getSequence(),aProtSeq, true, e.getRegExDigestion());
				}				
				if(terminalPeptUniqueness){
					currentPeptUniqueness=true;
				}
			}else{//'regular' peptide
				cleavageSiteDepUniqueness=determinePeptideUniquenessMod(aPept, aProtSeq, e.getRegExDigestion());
				if (!cleavageSiteDepUniqueness){
					regularPeptUniqueness=false;
				}
				if(regularPeptUniqueness){
					currentPeptUniqueness=true;
				}
			}
			if(currentPeptUniqueness){
				anOtherProtHasUniquePepts=true;
			}
		}
		if(aProtHasUniquePepts && anOtherProtHasUniquePepts){//sufficient requirement for dissimilarity is existence of at least on unique peptide in a protein
			isSimilar=false;
		}
		return isSimilar;
	}

	/**
	 * The method calculates the C terminal peptide based on the protein sequence and the specified regular expression.
	 * No modifications of the cleavage sites are considered.
	 * @param protSeq
	 * @param regExDigest
	 * @return the sequence of the C terminal peptide.
	 */
	private static String calcC_TerminalPeptide(String protSeq, String regExDigest){
		String cTermPept;
		String[] res=protSeq.split(regExDigest);
		cTermPept=res[res.length-1];
		return cTermPept;
	}
	
	/**
	 * The method calculates the N terminal peptide based on the protein sequence and the specified regular expression.
	 * No modifications of the cleavage sites are considered.
	 * @param protSeq
	 * @param regExDigest
	 * @return the sequence of the N terminal peptide.
	 */
	private static String calcN_TerminalPeptide(String protSeq, String regExDigest){
		String nTermPept;
		String[] res=protSeq.split(regExDigest);
		nTermPept=res[0];
		return nTermPept;
	}	
	
	/**
	 * The method calculates whether the given terminal peptide sequence is included as a terminal sequence in the given protein sequence.
	 * If this condition is not satisfied the peptide sequence is unique considering the bearing protein sequence. In this method variant it is assumed
	 * that no modifications can occur at the cleavage sites
	 * @param peptSeq
	 * @param protSeq
	 * @param isNTermPept
	 * @param regExDigest
	 * @return a boolean indicating whether a given terminal peptide sequence of a protein is unique regarding the comparison with another protein sequence.
	 */
	private static boolean determineUniquenessOfTerminalPeptide(String peptSeq, String protSeq, boolean isNTermPept, String regExDigest){
		boolean isUnique=false;
		String[] pepts= protSeq.split(regExDigest);
		if(isNTermPept){
			if(!peptSeq.equals(pepts[0])){
				isUnique=true;
			}
		}else{
			//System.out.println(String.valueOf(pepts.length));
			if(!peptSeq.equals(pepts[pepts.length-1])){
				isUnique=true;
			}
		}
		return isUnique;
	}	
	
	
	/**
	 * The method calculates the uniqueness of a certain peptide with respect to a compared protein sequence. 
	 * In this method variant modifications of cleavage sites are considered. 
	 * @param currPept
	 * @param anOtherProtSeq
	 * @param aCS
	 * @param performsN_terminalCleavage
	 * @param orientationModNotat
	 * @param regExDigest
	 * @param regExMod
	 * @return a boolean indicating whether a given terminal peptide sequence of a protein is unique regarding the comparison with another protein sequence. Post translational modifications are also considered.
	 */
	private static boolean determinePeptideUniquenessMod(Peptide currPept, String anOtherProtSeq, String regExDigest){
		boolean currPeptUniqueness=true;
		String aPeptSeqMod = currPept.getSequence();
		String[] pepts = anOtherProtSeq.split(regExDigest);
		 /*sufficient criteria for the existing uniqueness of currPept is no match with the non terminal peptides obtained from
		  * digestion of anOtherProtSeq. It is obvious that there is a cleavage site existing in the adjacent place of the protein 
		  * sequence (either N-terminal for C-terminal digested peptides or C-Terminal for N-terminal digested peptides.*/
		for (int i=1;i<pepts.length-1;i++){
			if(aPeptSeqMod.equals(pepts[i])){
				currPeptUniqueness=false;
			}
		}
		return currPeptUniqueness;
	}

	/**
	 * The method calculates the uniqueness of a certain peptide with respect to a compared protein sequence. 
	 * In this method variant it is assumed that cleavage sites (or other amino acids) are not modified. 
	 * @param currPept
	 * @param anOtherProtSeq
	 * @param aCS
	 * @param performsN_terminalCleavage
	 * @param currentPeptUniqueness_
	 * @return a boolean indicating whether a given peptide sequence of a protein is unique regarding the comparison with the peptides of another protein sequence.
	 */
	private static boolean determinePeptideUniqueness(Peptide currPept, String anOtherProtSeq, CleavageSite aCS, 
			boolean performsN_terminalCleavage){
		boolean currentPeptUniqueness=false;
		String aPeptSeqMod = currPept.getSequence();
		if(performsN_terminalCleavage){
			aPeptSeqMod=aPeptSeqMod+aCS.getAa_site();
		}else{
			aPeptSeqMod=aCS.getAa_site()+aPeptSeqMod;
		}
		if(!anOtherProtSeq.contains(aPeptSeqMod)){
			currentPeptUniqueness=true;			
		}
		return currentPeptUniqueness;
	}
}
